"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.compile = void 0;
const gq = __importStar(require("graphql"));
const fs = __importStar(require("fs/promises"));
const preamble_1 = require("./preamble");
const postamble_1 = require("./postamble");
function compile(args) {
    return __awaiter(this, void 0, void 0, function* () {
        const schemaData = yield fs.readFile(args.schema, 'utf8');
        let outputScript = '';
        const write = (s) => {
            outputScript += s + '\n';
        };
        let res = gq.parse(schemaData);
        const atomicTypes = new Map(res.definitions
            .flatMap(def => {
            if (def.kind === gq.Kind.SCALAR_TYPE_DEFINITION)
                return [[def.name.value, 'string']];
            if (def.kind === gq.Kind.ENUM_TYPE_DEFINITION)
                return [[def.name.value, def.name.value]];
            return [];
        })
            .concat([
            ['Int', 'number'],
            ['Float', 'number'],
            ['ID', 'string'],
            ['String', 'string'],
            ['Boolean', 'boolean'],
        ]));
        const inheritanceMap = new Map(res.definitions.flatMap(def => {
            var _a;
            if (def.kind === gq.Kind.OBJECT_TYPE_DEFINITION) {
                return [[def.name.value, (_a = def.interfaces) === null || _a === void 0 ? void 0 : _a.map(ifc => ifc.name.value)]];
            }
            return [];
        }));
        function isAtomic(typeName) {
            return !!atomicTypes.get(typeName);
        }
        function toTSType(scalar) {
            var _a;
            return (_a = atomicTypes.get(scalar)) !== null && _a !== void 0 ? _a : scalar;
        }
        function printAtomicTypes() {
            return `type $Atomic = ${Array.from(new Set(atomicTypes.values())).join(' | ')}
`;
        }
        function printTypeWrapped(wrappedType, wrapperDef, notNull = false) {
            switch (wrapperDef.kind) {
                case gq.Kind.NON_NULL_TYPE:
                    return `${printTypeWrapped(wrappedType, wrapperDef.type, true)}`;
                case gq.Kind.LIST_TYPE:
                    return `Array<${printTypeWrapped(wrappedType, wrapperDef.type)}>${!notNull ? ' | undefined' : ''}`;
                case gq.Kind.NAMED_TYPE:
                    return `${toTSType(wrappedType)}${!notNull ? ' | undefined' : ''}`;
            }
        }
        function printType(def, notNull = false) {
            switch (def.kind) {
                case gq.Kind.NON_NULL_TYPE:
                    return `${printType(def.type, true)}`;
                case gq.Kind.LIST_TYPE:
                    return `Array<${printType(def.type)}>${!notNull ? ' | undefined' : ''}`;
                case gq.Kind.NAMED_TYPE:
                    return `${toTSType(def.name.value)}${!notNull ? ' | undefined' : ''}`;
            }
        }
        function printTypeGql(def, notNull = false) {
            switch (def.kind) {
                case gq.Kind.NON_NULL_TYPE:
                    return `${printTypeGql(def.type, true)}`;
                case gq.Kind.LIST_TYPE:
                    return `[${printTypeGql(def.type)}]${notNull ? '!' : ''}`;
                case gq.Kind.NAMED_TYPE:
                    return `${def.name.value}${notNull ? '!' : ''}`;
            }
        }
        function printTypeBase(def) {
            switch (def.kind) {
                case gq.Kind.NON_NULL_TYPE:
                    return `${printTypeBase(def.type)}`;
                case gq.Kind.LIST_TYPE:
                    return `${printTypeBase(def.type)}`;
                case gq.Kind.NAMED_TYPE:
                    return `${def.name.value}`;
            }
        }
        function printInputField(def) {
            const maybe = def.type.kind !== gq.Kind.NON_NULL_TYPE ? '?' : '';
            return `${def.name.value}${maybe}: ${printType(def.type)}`;
        }
        function printDocumentation(description) {
            var _a;
            return ((_a = description === null || description === void 0 ? void 0 : description.value.length) !== null && _a !== void 0 ? _a : 0 > 0)
                ? `
/**
 * ${description === null || description === void 0 ? void 0 : description.value}
 */`
                : '';
        }
        function printObjectType(def) {
            var _a;
            const className = def.name.value;
            return `
${printDocumentation(def.description)}
export class ${className} extends $Base<"${className}"> {
  constructor() {
    super("${className}")
  }

  ${(_a = def.fields) === null || _a === void 0 ? void 0 : _a.map(f => printField(f, className)).join('\n')}
}`;
        }
        function printField(field, parentName) {
            var _a, _b, _c;
            const methodArgs = [];
            // const fieldType = printType(field.type)
            const fieldTypeName = printTypeBase(field.type);
            let hasArgs = false, hasSelector = false;
            if ((_a = field.arguments) === null || _a === void 0 ? void 0 : _a.length) {
                hasArgs = true;
                methodArgs.push(`args: Args`);
            }
            if (!isAtomic(fieldTypeName)) {
                hasSelector = true;
                methodArgs.push(`selectorFn: (s: ${fieldTypeName}) => [...Sel]`);
            }
            if (methodArgs.length > 0) {
                const argsType = `{
        ${((_b = field.arguments) !== null && _b !== void 0 ? _b : []).map(arg => printInputField(arg)).join('\n')},
      }`;
                const generics = (hasArgs ? [`Args extends VariabledInput<${argsType}>`] : []).concat(hasSelector ? [`Sel extends Selection<${fieldTypeName}>`] : []);
                return `
      ${printDocumentation(field.description)}
      ${field.name.value}<${generics.join(',')}>(${methodArgs.join(', ')}):$Field<"${field.name.value}", ${hasSelector ? printTypeWrapped('GetOutput<Sel>', field.type) : printType(field.type)} ${hasArgs ? `, GetVariables<${hasSelector ? 'Sel' : '[]'}, Args>` : ''}> {
      const options = {
        ${hasArgs
                    ? `argTypes: {
              ${(_c = field.arguments) === null || _c === void 0 ? void 0 : _c.map(arg => `${arg.name.value}: "${printTypeGql(arg.type)}"`).join(',\n')}
            },`
                    : ''}
        ${hasArgs ? `args,` : ''}

        ${hasSelector ? `selection: selectorFn(new ${fieldTypeName})` : ''}
      };
      return this.$_select("${field.name.value}", options) as any
    }
  `;
            }
            else {
                return `
      ${printDocumentation(field.description)}
      get ${field.name.value}(): $Field<"${field.name.value}", ${printType(field.type)}>  {
       return this.$_select("${field.name.value}") as any
      }`;
            }
        }
        function printInterface(def) {
            var _a;
            const className = def.name.value;
            return `
${printDocumentation(def.description)}
export class ${className} extends $Base<"${className}"> {
  constructor() {
    super("${className}")
  }
  ${(_a = def.fields) === null || _a === void 0 ? void 0 : _a.map(f => printField(f, className)).join('\n')}
}`;
        }
        function printInputObjectType(def) {
            var _a;
            return `
${printDocumentation(def.description)}
export type ${def.name.value} = {
  ${(_a = def.fields) === null || _a === void 0 ? void 0 : _a.map(field => printInputField(field)).join(',\n')}
}
    `;
        }
        function printInputTypeMap(defs) {
            return `
const $InputTypes: {[key: string]: {[key: string]: string}} = {
  ${defs
                .map(def => {
                var _a;
                return `  ${def.name.value}: {
    ${(_a = def.fields) === null || _a === void 0 ? void 0 : _a.map(field => `${field.name.value}: "${printTypeGql(field.type)}"`).join(',\n')}
  }`;
            })
                .join(',\n')}
}
`;
        }
        function printScalar(def) {
            return `
${printDocumentation(def.description)}
export type ${def.name.value} = unknown
`;
        }
        function printUnion(def) {
            var _a, _b;
            // TODO: collect all interfaces that the named type implements too
            const baseTypes = (_b = (_a = def.types) === null || _a === void 0 ? void 0 : _a.map(t => printTypeBase(t))) !== null && _b !== void 0 ? _b : [];
            const additionalTypes = Array.from(new Set(baseTypes.concat(baseTypes.flatMap(bt => { var _a; return (_a = inheritanceMap.get(bt)) !== null && _a !== void 0 ? _a : []; }))));
            const UnionObject = `{${additionalTypes.map(t => `${t}: ${t}`)}}`;
            return `
${printDocumentation(def.description)}
export class ${def.name.value} extends $Union<${UnionObject}, "${def.name.value}"> {
  constructor() {
    super(${UnionObject})
  }
}`;
        }
        function printEnumValue(def) {
            return `${printDocumentation(def.description)}
  ${def.name.value} = "${def.name.value}"`;
        }
        function printEnum(def) {
            var _a;
            return `
  ${printDocumentation(def.description)}
export enum ${def.name.value} {
  ${(_a = def.values) === null || _a === void 0 ? void 0 : _a.map(printEnumValue).join(',\n')}
}
  `;
        }
        function printSchema(def) {
            return `
  const $Root = {
    ${def.operationTypes.map(op => `${op.operation}: ${printType(op.type, true)}`).join(',\n')}
  }

  namespace $RootTypes {
    ${def.operationTypes
                .map(op => `export type ${op.operation} = ${printType(op.type, true)}`)
                .join('\n')}
  }
  `;
        }
        write(preamble_1.Preamble);
        write(printAtomicTypes());
        let rootNode = null;
        for (let def of res.definitions) {
            switch (def.kind) {
                case gq.Kind.OBJECT_TYPE_DEFINITION:
                    // the interfaces this object implements are def.interfaces
                    write(printObjectType(def));
                    break;
                case gq.Kind.INPUT_OBJECT_TYPE_DEFINITION:
                    write(printInputObjectType(def));
                    break;
                case gq.Kind.SCALAR_TYPE_DEFINITION:
                    write(printScalar(def));
                    break;
                case gq.Kind.UNION_TYPE_DEFINITION:
                    write(printUnion(def));
                    break;
                case gq.Kind.ENUM_TYPE_DEFINITION:
                    write(printEnum(def));
                    break;
                case gq.Kind.INTERFACE_TYPE_DEFINITION:
                    write(printInterface(def));
                    break;
                case gq.Kind.SCHEMA_DEFINITION:
                    rootNode = def;
            }
        }
        if (rootNode) {
            write(printSchema(rootNode));
            write((0, postamble_1.postamble)(rootNode.operationTypes.map(o => o.operation.toString())));
        }
        else {
            throw new Error('Schema does not have a root node');
        }
        write(printInputTypeMap(res.definitions.filter(def => def.kind === gq.Kind.INPUT_OBJECT_TYPE_DEFINITION)));
        yield fs.writeFile(args.output, outputScript);
    });
}
exports.compile = compile;
